@@ -206,6 +206,8 @@ static void          destroy_source  (GConfSource* source);
 
 static void          clear_cache     (GConfSource* source);
 
+static void          blow_away_locks (const char *address);
+
 static GConfBackendVTable xml_vtable = {
   x_shutdown,
   resolve_address,
@@ -224,13 +226,14 @@ static GConfBackendVTable xml_vtable = {
   set_schema,
   sync_all,
   destroy_source,
-  clear_cache
+  clear_cache,
+  blow_away_locks
 };
 
 static void          
 x_shutdown (GError** err)
 {
-  gconf_log(GCL_INFO, _("Unloading XML backend module."));
+  gconf_log(GCL_DEBUG, _("Unloading XML backend module."));
 }
 
 static void
@@ -267,86 +270,136 @@ writable (GConfSource* source,
   return TRUE;
 }
 
-static GConfSource*  
-resolve_address (const gchar* address, GError** err)
+static char*
+get_dir_from_address (const char *address,
+                      GError    **err)
 {
-  gchar* root_dir;
-  XMLSource* xsource;
-  GConfSource* source;
-  guint len;
-  gint flags = 0;
-  GConfLock* lock = NULL;
-  guint dir_mode = 0700;
-  guint file_mode = 0600;
+  char *root_dir;
+  int len;
   
-  root_dir = gconf_address_resource(address);
+  root_dir = gconf_address_resource (address);
 
   if (root_dir == NULL)
     {
-      gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, _("Couldn't find the XML root directory in the address `%s'"), address);
+      gconf_set_error (err, GCONF_ERROR_BAD_ADDRESS,
+                       _("Couldn't find the XML root directory in the address `%s'"),
+                       address);
       return NULL;
     }
 
   /* Chop trailing '/' to canonicalize */
-  len = strlen(root_dir);
+  len = strlen (root_dir);
 
   if (root_dir[len-1] == '/')
     root_dir[len-1] = '\0';
 
-  if (mkdir(root_dir, dir_mode) < 0)
+  return root_dir;
+}
+
+static char*
+get_lock_dir_from_root_dir (const char *root_dir)
+{
+  gchar* lockdir;
+  
+  lockdir = gconf_concat_dir_and_key (root_dir, "%gconf-xml-backend.lock");
+
+  return lockdir;
+}
+
+static GConfSource*  
+resolve_address (const gchar* address, GError** err)
+{
+  struct stat statbuf;
+  gchar* root_dir;
+  XMLSource* xsource;
+  GConfSource* source;
+  gint flags = 0;
+  GConfLock* lock = NULL;
+  guint dir_mode = 0700;
+  guint file_mode = 0600;
+  gchar** address_flags;
+  gchar** iter;
+  gboolean force_readonly;
+  
+  root_dir = get_dir_from_address (address, err);
+  if (root_dir == NULL)
+    return NULL;
+
+  if (stat (root_dir, &statbuf) == 0)
     {
-      if (errno != EEXIST)
+      /* Already exists, base our dir_mode on it */
+      dir_mode = _gconf_mode_t_to_mode (statbuf.st_mode);
+
+      /* dir_mode without search bits */
+      file_mode = dir_mode & (~0111); 
+    }
+  else if (mkdir (root_dir, dir_mode) < 0)
+    {
+      /* Error out even on EEXIST - shouldn't happen anyway */
+      gconf_set_error (err, GCONF_ERROR_FAILED,
+		       _("Could not make directory `%s': %s"),
+		      (gchar *)root_dir, g_strerror (errno));
+      g_free (root_dir);
+      return NULL;
+    }
+
+  force_readonly = FALSE;
+  
+  address_flags = gconf_address_flags (address);  
+  if (address_flags)
+    {
+      iter = address_flags;
+      while (*iter)
         {
-          gconf_set_error(err, GCONF_ERROR_FAILED,
-                          _("Could not make directory `%s': %s"),
-                          (gchar*)root_dir, strerror(errno));
-          g_free(root_dir);
-          return NULL;
-        }
-      else
-        {
-          /* Already exists, base our dir_mode on it */
-          struct stat statbuf;
-          if (stat(root_dir, &statbuf) == 0)
+          if (strcmp (*iter, "readonly") == 0)
             {
-              dir_mode = mode_t_to_mode(statbuf.st_mode);
-              /* dir_mode without search bits */
-              file_mode = dir_mode & (~0111); 
+              force_readonly = TRUE;
+              break;
             }
+
+          ++iter;
         }
     }
-  
+
+  g_strfreev (address_flags);
+
   {
     /* See if we're writable */
-    gboolean writable = FALSE;
+    gboolean writable;
     int fd;
     gchar* testfile;
 
-    testfile = g_strconcat(root_dir, "/.testing.writeability", NULL);    
+    writable = FALSE;
     
-    fd = open(testfile, O_CREAT|O_WRONLY, S_IRWXU);
-
-    if (fd >= 0)
+    if (!force_readonly)
       {
-        writable = TRUE;
-        close(fd);
-      }
+        testfile = g_strconcat(root_dir, "/.testing.writeability", NULL);    
         
-    unlink(testfile);
-
-    g_free(testfile);
+        fd = open(testfile, O_CREAT|O_WRONLY, S_IRWXU);
+        
+        if (fd >= 0)
+          {
+            writable = TRUE;
+            close(fd);
+          }
+        
+        unlink(testfile);
+        
+        g_free(testfile);
+      }
     
     if (writable)
       flags |= GCONF_SOURCE_ALL_WRITEABLE;
 
     /* We only do locking if it's writable,
-       which is sort of broken but close enough
-    */
-    if (writable)
+     * and if not using local locks,
+     * which is sort of broken but close enough
+     */
+    if (writable && !gconf_use_local_locks ())
       {
         gchar* lockdir;
 
-        lockdir = gconf_concat_dir_and_key(root_dir, "%gconf-xml-backend.lock");
+        lockdir = get_lock_dir_from_root_dir (root_dir);
         
         lock = gconf_get_lock(lockdir, err);
 
@@ -383,7 +436,7 @@ resolve_address (const gchar* address, GError** err)
   if (!(flags & GCONF_SOURCE_ALL_READABLE) &&
       !(flags & GCONF_SOURCE_ALL_WRITEABLE))
     {
-      gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, _("Can't read from or write to the XML root directory in the address `%s'"), address);
+      gconf_set_error(err, GCONF_ERROR_BAD_ADDRESS, _("Can't read from or write to the XML root directory in the address \"%s\""), address);
       g_free(root_dir);
       return NULL;
     }  
@@ -392,7 +445,7 @@ resolve_address (const gchar* address, GError** err)
 
   xsource = xs_new(root_dir, dir_mode, file_mode, lock);
 
-  gconf_log(GCL_INFO,
+  gconf_log(GCL_DEBUG,
             _("Directory/file permissions for XML source at root %s are: %o/%o"),
             root_dir, dir_mode, file_mode);
   
@@ -429,7 +482,7 @@ query_value (GConfSource* source,
      problem, since some errors may be added that need reporting. */
   if (error != NULL)
     {
-      gconf_log(GCL_WARNING, error->message);
+      gconf_log(GCL_WARNING, "%s", error->message);
       g_error_free(error);
       error = NULL;
     }
@@ -449,7 +502,7 @@ query_value (GConfSource* source,
       /* perhaps we should be reporting this error... */
       if (error != NULL)
         {
-          gconf_log(GCL_WARNING, error->message);
+          gconf_log(GCL_WARNING, "%s", error->message);
           g_error_free(error);
           error = NULL;
         }
@@ -477,7 +530,13 @@ query_metainfo  (GConfSource* source, const gchar* key,
       parent = NULL;
       
       if (dir != NULL)
-        return dir_get_metainfo(dir, key, err);
+        {
+          const gchar* relative_key;
+      
+          relative_key = gconf_key_key (key);
+
+          return dir_get_metainfo (dir, relative_key, err);
+        }
     }
 
   /* No metainfo found */
@@ -544,7 +603,22 @@ all_subdirs     (GConfSource* source,
 {
   Dir* dir;
   XMLSource* xs = (XMLSource*)source;
-
+  GError *sync_err;
+  
+  /* We have to sync before we can do this, to see which
+   * subdirs have gone away.
+   */
+  sync_err = NULL;
+  cache_sync (xs->cache, &sync_err);
+  if (sync_err)
+    {
+      gconf_log (GCL_WARNING, _("Error syncing the XML backend directory cache: %s"),
+                 sync_err->message);
+      g_error_free (sync_err);
+      sync_err = NULL;
+      /* continue, may as well try our best. */
+    }
+  
   dir = cache_lookup (xs->cache, key, FALSE, err);
   
   if (dir == NULL)
@@ -601,39 +675,32 @@ remove_dir      (GConfSource* source,
                  const gchar* key,
                  GError** err)
 {
-  XMLSource* xs = (XMLSource*)source;
-  Dir* dir;
-  
-  dir = cache_lookup(xs->cache, key, FALSE, err);
-  
-  if (dir == NULL)
-    return;
-  else
-    {
-      dir_mark_deleted(dir);
-    }
+  g_set_error (err, GCONF_ERROR,
+               GCONF_ERROR_FAILED,
+               _("Remove directory operation is no longer supported, just remove all the values in the directory"));
 }
 
 static void          
-set_schema      (GConfSource* source,
-                 const gchar* key,
-                 const gchar* schema_key,
-                 GError** err)
+set_schema (GConfSource *source,
+            const gchar *key,
+            const gchar *schema_key,
+            GError     **err)
 {
   XMLSource* xs = (XMLSource*)source;
 
   Dir* dir;
   gchar* parent;
-  
-  g_return_if_fail(schema_key != NULL);
 
-  parent = gconf_key_directory(key);
+  g_return_if_fail (source != NULL);
+  g_return_if_fail (key != NULL);
+
+  parent = gconf_key_directory (key);
   
-  g_assert(parent != NULL);
+  g_assert (parent != NULL);
   
-  dir = cache_lookup(xs->cache, parent, TRUE, err);
+  dir = cache_lookup (xs->cache, parent, TRUE, err);
   
-  g_free(parent);
+  g_free (parent);
   parent = NULL;
 
   if (dir == NULL)
@@ -642,9 +709,9 @@ set_schema      (GConfSource* source,
     {
       const gchar* relative_key;
       
-      relative_key = gconf_key_key(key);
+      relative_key = gconf_key_key (key);
       
-      dir_set_schema(dir, relative_key, schema_key, err);
+      dir_set_schema (dir, relative_key, schema_key, err);
     }
 }
 
@@ -654,7 +721,7 @@ sync_all        (GConfSource* source,
 {
   XMLSource* xs = (XMLSource*)source;
 
-  return cache_sync(xs->cache, err);
+  return cache_sync (xs->cache, err);
 }
 
 static void          
@@ -672,12 +739,70 @@ clear_cache     (GConfSource* source)
   cache_clean(xs->cache, 0);
 }
 
+static void
+blow_away_locks (const char *address)
+{
+  char *root_dir;
+  char *lock_dir;
+  DIR *dp;
+  struct dirent *dent;
+
+  /* /tmp locks should never be stuck, and possible security issue to
+   * blow them away
+   */
+  if (gconf_use_local_locks ())
+    return;
+  
+  root_dir = get_dir_from_address (address, NULL);
+  if (root_dir == NULL)
+    return;
+
+  lock_dir = get_lock_dir_from_root_dir (root_dir);
+
+  dp = opendir (lock_dir);
+  
+  if (dp == NULL)
+    {
+      g_printerr (_("Could not open lock directory for %s to remove locks: %s\n"),
+                  address, g_strerror (errno));
+      goto out;
+    }
+  
+  while ((dent = readdir (dp)) != NULL)
+    {
+      char *path;
+      
+      /* ignore ., .. (and any ..foo as an intentional who-cares bug) */
+      if (dent->d_name[0] == '.' &&
+          (dent->d_name[1] == '\0' || dent->d_name[1] == '.'))
+        continue;
+
+      path = g_build_filename (lock_dir, dent->d_name, NULL);
+
+      if (unlink (path) < 0)
+        {
+          g_printerr (_("Could not remove file %s: %s\n"),
+                      path, g_strerror (errno));
+        }
+
+      g_free (path);
+    }
+
+ out:
+
+  if (dp)
+    closedir (dp);
+  
+  g_free (root_dir);
+  g_free (lock_dir);
+}
+
 /* Initializer */
 
 G_MODULE_EXPORT const gchar*
 g_module_check_init (GModule *module)
 {
-  gconf_log(GCL_INFO, _("Initializing XML backend module"));
+  gconf_log(GCL_DEBUG, _("Initializing XML backend module"));
 
   LIBXML_TEST_VERSION;
   xmlKeepBlanksDefault(1);
@@ -720,7 +845,7 @@ xs_new       (const gchar* root_dir, guint dir_mode, guint file_mode, GConfLock*
 
   xs->root_dir = g_strdup(root_dir);
 
-  xs->cache = cache_new(xs->root_dir, dir_mode, file_mode);
+  xs->cache = cache_get(xs->root_dir, dir_mode, file_mode);
 
   xs->timeout_id = g_timeout_add(1000*60*5, /* 1 sec * 60 s/min * 5 min */
                                  cleanup_timeout,
@@ -745,8 +870,8 @@ xs_destroy   (XMLSource* xs)
      situation */
   if (xs->lock != NULL && !gconf_release_lock(xs->lock, &error))
     {
-      gconf_log(GCL_ERR, _("Failed to give up lock on XML dir `%s': %s"),
-                xs->root_dir, error->message);
+      gconf_log (GCL_ERR, _("Failed to give up lock on XML directory \"%s\": %s"),
+                 xs->root_dir, error->message);
       g_error_free(error);
       error = NULL;
     }
@@ -757,7 +882,8 @@ xs_destroy   (XMLSource* xs)
       gconf_log(GCL_ERR, "timeout not found to remove?");
     }
   
-  cache_destroy(xs->cache);
+  cache_unref(xs->cache);
   g_free(xs->root_dir);
   g_free(xs);
 }
+
